

cdef extern from *:
    """
    //preprocessor creates needed struct-type and all function definitions

    //set with keys of type int64 -> resulting typename: kh_int64set_t;
    KHASH_INIT(int64set, int64_t, char, 0, kh_int64_hash_func, kh_int64_hash_equal)

    //Generated code:
    /*
        typedef struct kh_int64set_s { 
            khint_t n_buckets, size, n_occupied, upper_bound; 
            khint32_t *flags; 
            int64_t *keys; 
            char *vals; 
        } kh_int64set_t; 



        static inline __attribute__ ((__unused__)) kh_int64set_t *kh_init_int64set(void) { 
            return (kh_int64set_t*)calloc(1,sizeof(kh_int64set_t)); 
        } 


        static inline __attribute__ ((__unused__)) void kh_destroy_int64set(kh_int64set_t *h) { 
            if (h) { 
                free((void *)h->keys); 
                free(h->flags); 
                free((void *)h->vals); 
                free(h); 
            } 
        } 


        static inline __attribute__ ((__unused__)) void kh_clear_int64set(kh_int64set_t *h) { 
            if (h && h->flags) { 
                memset(h->flags, 0xaa, ((h->n_buckets) < 16? 1 : (h->n_buckets)>>4) * sizeof(khint32_t)); h->size = h->n_occupied = 0; 
            }
        } 



        static inline __attribute__ ((__unused__)) khint_t kh_get_int64set(const kh_int64set_t *h, khint64_t key) { 
            if (h->n_buckets) { 
                khint_t k, i, last, mask, step = 0; 
                mask = h->n_buckets - 1; 
                k = (khint32_t)((key)>>33^(key)^(key)<<11); 
                i = k & mask; 
                last = i; 
                while (!((h->flags[i>>4]>>((i&0xfU)<<1))&2) && (((h->flags[i>>4]>>((i&0xfU)<<1))&1) || !((h->keys[i]) == (key)))) { 
                    i = (i + (++step)) & mask; 
                    if (i == last) 
                        return h->n_buckets; 
                } 
                return ((h->flags[i>>4]>>((i&0xfU)<<1))&3)? h->n_buckets : i; 
            } 
            else return 0; 
        } 

        S

        static inline __attribute__ ((__unused__)) int kh_resize_int64set(kh_int64set_t *h, khint_t new_n_buckets) { 
            khint32_t *new_flags = 0;
            khint_t j = 1; 
            { 
                (--(new_n_buckets), (new_n_buckets)|=(new_n_buckets)>>1, (new_n_buckets)|=(new_n_buckets)>>2, (new_n_buckets)|=(new_n_buckets)>>4, (new_n_buckets)|=(new_n_buckets)>>8, (new_n_buckets)|=(new_n_buckets)>>16, ++(new_n_buckets)); 
                if (new_n_buckets < 4) 
                    new_n_buckets = 4; 
                if (h->size >= (khint_t)
                    (new_n_buckets * __ac_HASH_UPPER + 0.5)) j = 0; 
                else { 
                    new_flags = (khint32_t*)malloc(((new_n_buckets) < 16? 1 : (new_n_buckets)>>4) * sizeof(khint32_t)); 
                    if (!new_flags) 
                        return -1; 
                    memset(new_flags, 0xaa, ((new_n_buckets) < 16? 1 : (new_n_buckets)>>4) * sizeof(khint32_t)); 
                    if (h->n_buckets < new_n_buckets) { 
                        khint64_t *new_keys = (khint64_t*)realloc((void *)h->keys,new_n_buckets * sizeof(khint64_t)); 
                        if (!new_keys) { 
                            free(new_flags); 
                            return -1; 
                        } 
                        h->keys = new_keys; 
                        if (0) { 
                            char *new_vals = (char*)realloc((void *)h->vals,new_n_buckets * sizeof(char)); 
                            if (!new_vals) { 
                                free(new_flags); 
                                return -1; 
                            } 
                            h->vals = new_vals; 
                        } 
                    } 
                } 
            } 
            if (j) { 
                for (j = 0; j != h->n_buckets; ++j) { 
                    if (((h->flags[j>>4]>>((j&0xfU)<<1))&3) == 0) { 
                        khint64_t key = h->keys[j]; 
                        char val; 
                        khint_t new_mask; 
                        new_mask = new_n_buckets - 1; 
                        if (0) 
                            val = h->vals[j]; 
                        (h->flags[j>>4]|=1ul<<((j&0xfU)<<1)); 
                        while (1) { 
                            khint_t k, i, step = 0; 
                            k = (khint32_t)((key)>>33^(key)^(key)<<11); 
                            i = k & new_mask; 
                            while (!((new_flags[i>>4]>>((i&0xfU)<<1))&2)) 
                                i = (i + (++step)) & new_mask; 
                            (new_flags[i>>4]&=~(2ul<<((i&0xfU)<<1))); 
                            if (i < h->n_buckets && ((h->flags[i>>4]>>((i&0xfU)<<1))&3) == 0) {
                                { 
                                    khint64_t tmp = h->keys[i]; h->keys[i] = key; key = tmp; 
                                } 
                                if (0) { 
                                    char tmp = h->vals[i]; 
                                    h->vals[i] = val; 
                                    val = tmp; 
                                } 
                                (h->flags[i>>4]|=1ul<<((i&0xfU)<<1)); 
                            } 
                            else { 
                                h->keys[i] = key; 
                                if (0) 
                                    h->vals[i] = val; 
                                break; 
                            } 
                        } 
                    } 
                } 
                if (h->n_buckets > new_n_buckets) { 
                    h->keys = (khint64_t*)realloc((void *)h->keys,new_n_buckets * sizeof(khint64_t)); 
                    if (0) 
                        h->vals = (char*)realloc((void *)h->vals,new_n_buckets * sizeof(char)); 
                } 
                free(h->flags); 
                h->flags = new_flags; 
                h->n_buckets = new_n_buckets; 
                h->n_occupied = h->size; 
                h->upper_bound = (khint_t)(h->n_buckets * __ac_HASH_UPPER + 0.5); 
            } 
            return 0; 
        } 


        static inline __attribute__ ((__unused__)) khint_t kh_put_int64set(kh_int64set_t *h, khint64_t key, int *ret) { 
            khint_t x; 
            if (h->n_occupied >= h->upper_bound) { 
                if (h->n_buckets > (h->size<<1)) { 
                    if (kh_resize_int64set(h, h->n_buckets - 1) < 0) { 
                        *ret = -1; 
                        return h->n_buckets; 
                    } 
                } 
                else if (kh_resize_int64set(h, h->n_buckets + 1) < 0) { 
                    *ret = -1; 
                    return h->n_buckets; 
                } 
            } 
            { 
                khint_t k, i, site, last, mask = h->n_buckets - 1, step = 0; 
                x = site = h->n_buckets; 
                k = (khint32_t)((key)>>33^(key)^(key)<<11); 
                i = k & mask; 
                if (((h->flags[i>>4]>>((i&0xfU)<<1))&2)) 
                    x = i; 
                else { 
                    last = i; 
                    while (!((h->flags[i>>4]>>((i&0xfU)<<1))&2) && (((h->flags[i>>4]>>((i&0xfU)<<1))&1) || !((h->keys[i]) == (key)))) { 
                        if (((h->flags[i>>4]>>((i&0xfU)<<1))&1)) 
                            site = i; 
                        i = (i + (++step)) & mask; 
                        if (i == last) { 
                            x = site; 
                            break; 
                        } 
                    } 
                    if (x == h->n_buckets) { 
                        if (((h->flags[i>>4]>>((i&0xfU)<<1))&2) && site != h->n_buckets) 
                            x = site; 
                        else 
                            x = i; 
                    } 
                } 
            } 
            if (((h->flags[x>>4]>>((x&0xfU)<<1))&2)) { 
                h->keys[x] = key; 
                (h->flags[x>>4]&=~(3ul<<((x&0xfU)<<1))); 
                ++h->size; 
                ++h->n_occupied; 
                *ret = 1; 
            } 
            else if (((h->flags[x>>4]>>((x&0xfU)<<1))&1)) { 
                h->keys[x] = key; 
                (h->flags[x>>4]&=~(3ul<<((x&0xfU)<<1))); 
                ++h->size; 
                *ret = 2; 
            } 
            else 
                *ret = 0; 
            return x; 
        } 

        static inline __attribute__ ((__unused__)) void kh_del_int64set(kh_int64set_t *h, khint_t x) { 
            if (x != h->n_buckets && !((h->flags[x>>4]>>((x&0xfU)<<1))&3)) { 
                (h->flags[x>>4]|=1ul<<((x&0xfU)<<1)); 
                --h->size; 
            } 
        }
    */
    """
    pass




cdef extern from *:
    """
    // correct handling for float64/32 <-> int64/32
    #include <string.h>

    typedef double float64_t;
    typedef float float32_t;

    //don't pun and alias:

    inline int64_t f64_to_i64(float64_t val){
          int64_t res; 
          memcpy(&res, &val, sizeof(float64_t)); 
          return res;
    } 

    inline float64_t i64_to_f64(int64_t val){
          float64_t res; 
          memcpy(&res, &val, sizeof(float64_t)); 
          return res;
    }

    inline int32_t f32_to_i32(float32_t val){
          int32_t res; 
          memcpy(&res, &val, sizeof(float32_t)); 
          return res;
    } 

    inline float32_t i32_to_f32(int32_t val){
          float32_t res; 
          memcpy(&res, &val, sizeof(float32_t)); 
          return res;
    }

    // HASH AND EQUAL FUNCTIONS:
    // 64bit
    //khash has nothing predefined for float/double
    //      in the first stet we add needed functionality
    typedef float64_t khfloat64_t;

    //right for all but not -0.0 and NAN
    #define kh_float64_hash_func(key) (khint32_t)((f64_to_i64(key))>>33^(f64_to_i64(key))^(f64_to_i64(key))<<11)

    //right for all except NAN
    #define kh_float64_hash_func_0(key) ((key)==0.0 ? kh_float64_hash_func(0.0) : kh_float64_hash_func(key))

    //right for all, also 0.0 and NAN
    #define kh_float64_hash_func_0_NAN(key) ((key) != (key) ? kh_float64_hash_func_0(NAN) : kh_float64_hash_func(key))

    //                                                       take care of nans:
    #define kh_float64_hash_equal(a, b) ((a) == (b) || ((b) != (b) && (a) != (a)))

    // 32bit
    typedef float float32_t;
    typedef float32_t khfloat32_t;


    //right for all but not -0.0 and NAN
    #define kh_float32_hash_func(key) (khint32_t)(f32_to_i32(key))

    //right for all except NAN
    #define kh_float32_hash_func_0(key) ((key)==0.0f ? kh_float32_hash_func(0.0f) : kh_float32_hash_func(key))

    //right for all, also 0.0 and NAN
    #define kh_float32_hash_func_0_NAN(key) ((key) != (key) ? kh_float32_hash_func_0(NANF) : kh_float32_hash_func(key))

    //                                                       take care of nans:
    #define kh_float32_hash_equal(a, b) ((a) == (b) || ((b) != (b) && (a) != (a)))

    """  
    pass



from cpython cimport array


cdef extern from *:
    """
    // from float_utils.pxi
    """  
    int64_t   f64_to_i64(float64_t val)
    float64_t i64_to_f64(int64_t   val)
    int32_t   f32_to_i32(float32_t val)
    float32_t i32_to_f32(int32_t   val)


# different implementations:
include "sets/int64/int64set_impl.pxi"
include "sets/int32/int32set_impl.pxi"
include "sets/float64/float64set_impl.pxi"
include "sets/float32/float32set_impl.pxi"
include "sets/pyobject/pyobjectset_impl.pxi"



from cpython cimport array

cdef extern from *:
    """
    // from float_utils.pxi
    """  
    int64_t   f64_to_i64(float64_t val)
    float64_t i64_to_f64(int64_t   val)
    int32_t   f32_to_i32(float32_t val)
    float32_t i32_to_f32(int32_t   val)

# different implementations:
include "maps/int64/int64to64map_impl.pxi"
include "maps/int32/int32to32map_impl.pxi"
include "maps/float64/float64to64map_impl.pxi"
include "maps/float32/float32to32map_impl.pxi"
include "maps/pyobject/pyobjectmap_impl.pxi"




# see float_utils.pxi for definitions

cdef extern from *:
    """
    // preprocessor creates needed struct-type and all function definitions

    // map with keys of type float32 -> und result int32

    #define KHASH_MAP_INIT_FLOAT32(name, khval_t)								\
	    KHASH_INIT(name, khfloat32_t, khval_t, 1, kh_float32_hash_func, kh_float32_hash_equal)

    KHASH_MAP_INIT_FLOAT64(float32to32map, int32_t)

    """
    pass





#
#
# Don't edit it, unless this is I_n_t_6_4_to_6_4_m_a_p header
#
# run sh all_from_XXX.sh to create it from blueprint -  I_n_t_6_4_to_6_4_m_a_p

include "init_int32to32map.pxi"
cdef extern from *:

    ctypedef struct kh_int32to32map_t:
        khint_t n_buckets, size, n_occupied, upper_bound
        uint32_t *flags
        int32_t *keys
        int32_t *vals  

    kh_int32to32map_t* kh_init_int32to32map() nogil
    void kh_destroy_int32to32map(kh_int32to32map_t*) nogil
    void kh_clear_int32to32map(kh_int32to32map_t*) nogil
    khint_t kh_get_int32to32map(kh_int32to32map_t*, int32_t) nogil
    void kh_resize_int32to32map(kh_int32to32map_t*, khint_t) nogil
    khint_t kh_put_int32to32map(kh_int32to32map_t*, int32_t, int* result) nogil
    void kh_del_int32to32map(kh_int32to32map_t*, khint_t) nogil

    #specializing "kh_exist"-macro 
    bint kh_exist_int32to32map "kh_exist" (kh_int32to32map_t*, khint_t) nogil


cdef class Int32to32Map:
    cdef kh_int32to32map_t *table
    cdef bint for_int

    cdef bint contains(self, int32_t key) except *
    cdef Int32to32MapIterator get_iter(self)
    cdef khint_t size(self) 
    cpdef void put_int32(self, int32_t key, int32_t value) except *
    cpdef int32_t get_int32(self, int32_t key) except *
    cpdef void put_float32(self, int32_t key, float32_t value) except *
    cpdef float32_t get_float32(self, int32_t key) except *
    cpdef void discard(self, int32_t key) except *
    

cdef struct int32to32_key_val_pair:
    int32_t key
    int32_t val


cdef class Int32to32MapIterator:
    cdef khint_t   it
    cdef khint_t   size
    cdef Int32to32Map  parent

    cdef bint has_next(self) except *
    cdef int32to32_key_val_pair next(self) except *
    cdef void __move(self) except *




__version__=(1,0,2)

from .khashsets import Int64Set, Int64SetIterator, isin_int64,  Int64Set_from, Int64Set_from_buffer
from .khashsets import Int32Set, Int32SetIterator, isin_int32,  Int32Set_from, Int32Set_from_buffer
from .khashsets import Float64Set, Float64SetIterator, isin_float64,  Float64Set_from, Float64Set_from_buffer
from .khashsets import Float32Set, Float32SetIterator, isin_float32,  Float32Set_from, Float32Set_from_buffer
from .khashsets import PyObjectSet, PyObjectSetIterator, isin_pyobject,  PyObjectSet_from, PyObjectSet_from_buffer


from .khashmaps import Int64to64Map, Int64to64MapIterator, Int64to64Map_from_int64_buffer, Int64to64Map_from_float64_buffer 
from .khashmaps import Int32to32Map, Int32to32MapIterator, Int32to32Map_from_int32_buffer, Int32to32Map_from_float32_buffer 
from .khashmaps import Float64to64Map, Float64to64MapIterator, Float64to64Map_from_int64_buffer, Float64to64Map_from_float64_buffer 
from .khashmaps import Float32to32Map, Float32to32MapIterator, Float32to32Map_from_int32_buffer, Float32to32Map_from_float32_buffer 
from .khashmaps import PyObjectMap, PyObjectMapIterator, PyObjectMap_from_object_buffer


from .unique import unique_int64, unique_int32, unique_float64, unique_float32
from .unique import unique_stable_int64, unique_stable_int32, unique_stable_float64, unique_stable_float32



cdef extern from *:
    """
    //khash has nothing predefined for Pyobject
    #include <Python.h>

    typedef PyObject* pyobject_t;
    typedef pyobject_t khpyobject_t;


    inline int pyobject_cmp(PyObject* a, PyObject* b) {
	    int result = PyObject_RichCompareBool(a, b, Py_EQ);
	    if (result < 0) {
		    PyErr_Clear();
		    return 0;
	    }
        if (result == 0) {  // still could be two NaNs
            return PyFloat_CheckExact(a) &&
                   PyFloat_CheckExact(b) &&
                   Py_IS_NAN(PyFloat_AS_DOUBLE(a)) &&
                   Py_IS_NAN(PyFloat_AS_DOUBLE(b));
        }
	    return result;
    }

    inline khint32_t pyobject_hash(PyObject* key){
        khint64_t hash = PyObject_Hash(key);
        return (khint32_t) ((hash)>>33^(hash)^(hash)<<11);
    }

    // For PyObject_Hash holds:
    //    hash(0.0) == 0 == hash(-0.0)
    //    hash(X) == 0 if X is a NaN-value
    // so it is OK to use it directly
    
    #define kh_pyobject_hash_func(key) (pyobject_hash(key))
    #define kh_pyobject_hash_equal(a, b) (pyobject_cmp(a, b))
    """
    pass



#
#
#  Attention the problem with PyObject is reference counting
#
#

cdef extern from *:
    """
    //khash has nothing predefined for Pyobject

    /*! @function
      @abstract     Instantiate a hash set containing 64-bit integer keys
      @param  name  Name of the hash table [symbol]
     */
    #define KHASH_SET_INIT_PYOBJECT(name)										\
	    KHASH_INIT(name, khpyobject_t, char, 0, kh_pyobject_hash_func, kh_pyobject_hash_equal)

    //preprocessor creates needed struct-type and all function definitions 
    //set with keys of type pyobject -> resulting typename: kh_pyobjectset_t;
    KHASH_SET_INIT_PYOBJECT(pyobjectset)

    //TODO: add generated code:
    """
    pass






cdef extern from *:
    """
    // preprocessor creates needed struct-type and all function definitions

    // map with keys of type int64 -> und result int64
    // resulting typename: kh_int64to64map_t;
    KHASH_INIT(int64to64map, int64_t, int64_t, 1, kh_int64_hash_func, kh_int64_hash_equal)
  

    """
    pass





#
#
# Don't edit it, unless this is I_n_t_6_4_to_6_4_m_a_p header
#
# run sh all_from_XXX.sh to create it from blueprint -  I_n_t_6_4_to_6_4_m_a_p

include "init_int64to64map.pxi"
cdef extern from *:

    ctypedef struct kh_int64to64map_t:
        khint_t n_buckets, size, n_occupied, upper_bound
        uint32_t *flags
        int64_t *keys
        int64_t *vals  

    kh_int64to64map_t* kh_init_int64to64map() nogil
    void kh_destroy_int64to64map(kh_int64to64map_t*) nogil
    void kh_clear_int64to64map(kh_int64to64map_t*) nogil
    khint_t kh_get_int64to64map(kh_int64to64map_t*, int64_t) nogil
    void kh_resize_int64to64map(kh_int64to64map_t*, khint_t) nogil
    khint_t kh_put_int64to64map(kh_int64to64map_t*, int64_t, int* result) nogil
    void kh_del_int64to64map(kh_int64to64map_t*, khint_t) nogil

    #specializing "kh_exist"-macro 
    bint kh_exist_int64to64map "kh_exist" (kh_int64to64map_t*, khint_t) nogil


cdef class Int64to64Map:
    cdef kh_int64to64map_t *table
    cdef bint for_int

    cdef bint contains(self, int64_t key) except *
    cdef Int64to64MapIterator get_iter(self)
    cdef khint_t size(self) 
    cpdef void put_int64(self, int64_t key, int64_t value) except *
    cpdef int64_t get_int64(self, int64_t key) except *
    cpdef void put_float64(self, int64_t key, float64_t value) except *
    cpdef float64_t get_float64(self, int64_t key) except *
    cpdef void discard(self, int64_t key) except *
    

cdef struct int64to64_key_val_pair:
    int64_t key
    int64_t val


cdef class Int64to64MapIterator:
    cdef khint_t   it
    cdef khint_t   size
    cdef Int64to64Map  parent

    cdef bint has_next(self) except *
    cdef int64to64_key_val_pair next(self) except *
    cdef void __move(self) except *





from libc.stdint cimport int64_t,  int32_t
from .floatdef cimport float64_t, float32_t

cpdef unique_int64(int64_t[:] vals, double size_hint=*)
cpdef unique_int32(int32_t[:] vals, double size_hint=*)
cpdef unique_float64(float64_t[:] vals, double size_hint=*)
cpdef unique_float32(float32_t[:] vals, double size_hint=*)

cpdef unique_stable_int64(int64_t[:] vals, double size_hint=*)
cpdef unique_stable_int32(int32_t[:] vals, double size_hint=*)
cpdef unique_stable_float64(float64_t[:] vals, double size_hint=*)
cpdef unique_stable_float32(float32_t[:] vals, double size_hint=*)




from cpython.ref cimport Py_INCREF,Py_DECREF

cdef class PyObjectMap:

    def __cinit__(self, *, number_of_elements_hint=None):
        """
        number_of_elements_hint - number of elements without the need of reallocation.
        """
        self.table = kh_init_pyobjectmap()
        if number_of_elements_hint is not None:
            kh_resize_pyobjectmap(self.table, element_n_to_bucket_n(number_of_elements_hint))

    def __len__(self):
        return self.size()
  
    cdef khint_t size(self):
        return self.table.size
        

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self.table is not NULL:
            for i in range(self.table.size):
                if kh_exist_pyobjectmap(self.table, i):
                    Py_DECREF(<object>(self.table.keys[i]))
                    Py_DECREF(<object>(self.table.vals[i]))
            kh_destroy_pyobjectmap(self.table)
            self.table = NULL

    def __contains__(self, object key):
        return self.contains(key)


    cdef bint contains(self, object key) except *:
        cdef khint_t k
        k = kh_get_pyobjectmap(self.table, <pyobject_t>key)
        return k != self.table.n_buckets


    cpdef void put_object(self, object key, object val) except *:
        cdef:
            khint_t k
            int ret = 0
        k = kh_put_pyobjectmap(self.table, <pyobject_t>key, &ret)
        if not ret:
            Py_DECREF(<object>(self.table.vals[k]))
        else:
            Py_INCREF(key)
        Py_INCREF(val)
        self.table.vals[k] = <pyobject_t> val

    
    def __setitem__(self, key, val):
        self.put_object(key, val)

    cpdef object get_object(self, object key):
        k = kh_get_pyobjectmap(self.table, <pyobject_t>key)
        if k != self.table.n_buckets:
            return <object>self.table.vals[k]
        else:
            raise KeyError("No such key: "+str(key))

    def __getitem__(self, key):
        return self.get_object(key)

    
    cpdef void discard(self, object key) except *:
        cdef khint_t k
        k = kh_get_pyobjectmap(self.table, <pyobject_t>key)
        if k != self.table.n_buckets:
            Py_DECREF(<object>(self.table.keys[k]))
            Py_DECREF(<object>(self.table.vals[k]))
            kh_del_pyobjectmap(self.table, k)


    cdef PyObjectMapIterator get_iter(self):
        return PyObjectMapIterator(self)

    def __iter__(self):
        return self.get_iter()


### Iterator:
cdef class PyObjectMapIterator:

    cdef void __move(self) except *:
        while self.it<self.size and not kh_exist_pyobjectmap(self.parent.table, self.it):
              self.it+=1       

    cdef bint has_next(self) except *:
        return self.it != self.parent.table.n_buckets
        
    cdef pyobject_key_val_pair next(self) except *:
        cdef pyobject_key_val_pair result 
        result.key = self.parent.table.keys[self.it]
        result.val = self.parent.table.vals[self.it]
        self.it+=1#ensure at least one move!
        self.__move()
        return result


    def __cinit__(self, PyObjectMap parent):
        self.parent = parent
        self.size = parent.table.n_buckets
        #search the start:
        self.it = 0
        self.__move()

    def __next__(self):
        cdef pyobject_key_val_pair p
        if self.has_next():
            p = self.next()
            return {'key' : <object>p.key, 'val' : <object>p.val}
        else:
            raise StopIteration

### Utils:

def PyObjectMap_from_object_buffer(object[:] keys, object[:] vals, double size_hint=0.0):
    cdef Py_ssize_t n = len(keys)
    cdef Py_ssize_t b = len(vals)
    if b < n:
        n = b
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=PyObjectMap(number_of_elements_hint=at_least_needed)
    cdef Py_ssize_t i
    for i in range(n):
        res.put_object(keys[i], vals[i])
    return res

    





#
#
# Don't edit it, unless this is I_n_t_6_4_S_e_t implementation
#
# run sh all_from_XXX.sh to create it from bluepring - I_n_t_6_4_S_e_t
#
#
from cpython.ref cimport Py_INCREF,Py_DECREF

cdef class PyObjectSet:

    def __cinit__(self,  *, number_of_elements_hint=None):
        """
        number_of_elements_hint - number of elements without the need of reallocation.
        """
        self.table = kh_init_pyobjectset()
        if number_of_elements_hint is not None:
            kh_resize_pyobjectset(self.table, element_n_to_bucket_n(number_of_elements_hint))

    def __len__(self):
        return self.size()
  
    cdef khint_t size(self):
        return self.table.size
        

    def __dealloc__(self):
        cdef khint_t i = 0
        if self.table is not NULL:
            for i in range(self.table.size):
                if kh_exist_pyobjectset(self.table, i):
                    Py_DECREF(<object>self.table.keys[i])
            kh_destroy_pyobjectset(self.table)
            self.table = NULL

    def __contains__(self, object key):
        return self.contains(key)


    cdef bint contains(self, object key) except *:
        cdef khint_t k
        k = kh_get_pyobjectset(self.table, <pyobject_t>key)
        return k != self.table.n_buckets


    cpdef void add(self, object key) except *:
        cdef:
            khint_t k
            int ret = 0
            pyobject_t key_ptr = <pyobject_t> key
        k = kh_put_pyobjectset(self.table, key_ptr, &ret)
        if ret: 
            #element was really added, so we need to increase reference
            Py_INCREF(key)

    
    cpdef void discard(self, object key) except *:
        cdef khint_t k
        cdef pyobject_t key_ptr = <pyobject_t> key
        k = kh_get_pyobjectset(self.table, key_ptr)
        if k != self.table.n_buckets:
            Py_DECREF(<object>self.table.keys[k])
            kh_del_pyobjectset(self.table, k)


    cdef PyObjectSetIterator get_iter(self):
        return PyObjectSetIterator(self)

    def __iter__(self):
        return self.get_iter()

    def get_state_info(self):
        return {"n_buckets" : self.table.n_buckets, 
                "n_occupied" : self.table.n_occupied, 
                "upper_bound" : self.table.upper_bound}


### Iterator:
cdef class PyObjectSetIterator:

    cdef void __move(self) except *:
        while self.it<self.size and not kh_exist_pyobjectset(self.parent.table, self.it):
              self.it+=1       

    cdef bint has_next(self) except *:
        return self.it != self.parent.table.n_buckets
        
    cdef object next(self):
        cdef pyobject_t result = self.parent.table.keys[self.it]
        self.it+=1#ensure at least one move!
        self.__move()
        return <object>result


    def __cinit__(self, PyObjectSet parent):
        self.parent = parent
        self.size = parent.table.n_buckets
        #search the start:
        self.it = 0
        self.__move()

    def __next__(self):
        if self.has_next():
            return self.next()
        else:
            raise StopIteration

### Utils:

def PyObjectSet_from(it):
    res=PyObjectSet()
    for i in it:
        res.add(i)
    return res
    
def PyObjectSet_from_buffer(object[:] buf, double size_hint=0.0):
    cdef Py_ssize_t n = len(buf)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=PyObjectSet(number_of_elements_hint=at_least_needed)
    cdef Py_ssize_t i
    for i in range(n):
        res.add(buf[i])
    return res

from libc.stdint cimport  uint8_t

def isin_pyobject(object[:] query, PyObjectSet db, uint8_t[:] result):
    cdef size_t i
    cdef size_t n=len(query)
    for i in range(n):
        result[i]=db.contains(query[i])






# see float_utils.pxi for definitions

cdef extern from *:
    """
    // preprocessor creates needed struct-type and all function definitions

    // map with keys of type float64 -> und result int64

    #define KHASH_MAP_INIT_FLOAT64(name, khval_t)								\
	    KHASH_INIT(name, khfloat64_t, khval_t, 1, kh_float64_hash_func, kh_float64_hash_equal)

    KHASH_MAP_INIT_FLOAT64(float64to64map, int64_t)

    """
    pass






# see float_utils.pxi for definitions

cdef extern from *:
    """
    // preprocessor creates needed struct-type and all function definitions

    // map with keys of type pyobject -> result pyobject
    #define KHASH_MAP_INIT_PYOBJECT(name, khval_t)										\
	    KHASH_INIT(name, khpyobject_t, khval_t, 1, kh_pyobject_hash_func, kh_pyobject_hash_equal)

    //preprocessor creates needed struct-type and all function definitions 
    //set with keys of type pyobject -> resulting typename: kh_pyobjectmap_t;
    KHASH_MAP_INIT_PYOBJECT(pyobjectmap, pyobject_t)

    """
    pass






cdef class Float32to32Map:

    def __cinit__(self, *, number_of_elements_hint=None, for_int=True):
        """
        number_of_elements_hint - number of elements without the need of reallocation.
        for_int  if True, __setitem__/__getitem__ sets/gets a in32-object, otherwise a float32-object
        """
        self.for_int = for_int
        self.table = kh_init_float32to32map()
        if number_of_elements_hint is not None:
            kh_resize_float32to32map(self.table, element_n_to_bucket_n(number_of_elements_hint))

    def __len__(self):
        return self.size()
  
    cdef khint_t size(self):
        return self.table.size
        

    def __dealloc__(self):
        if self.table is not NULL:
            kh_destroy_float32to32map(self.table)
            self.table = NULL

    def __contains__(self, float32_t key):
        return self.contains(key)


    cdef bint contains(self, float32_t key) except *:
        cdef khint_t k
        k = kh_get_float32to32map(self.table, key)
        return k != self.table.n_buckets


    cpdef void put_int32(self, float32_t key, int32_t val) except *:
        cdef:
            khint_t k
            int ret = 0

        k = kh_put_float32to32map(self.table, key, &ret)
        self.table.keys[k] = key
        self.table.vals[k] = val

    cpdef void put_float32(self, float32_t key, float32_t val) except *:
        self.put_int32(key, f32_to_i32(val));

    
    def __setitem__(self, key, val):
        if self.for_int:
            self.put_int32(key, val)
        else:
            self.put_float32(key, val)

    cpdef int32_t get_int32(self, float32_t key) except *:
        k = kh_get_float32to32map(self.table, key)
        if k != self.table.n_buckets:
            return self.table.vals[k]
        else:
            raise KeyError("No such key: "+str(key))

    cpdef float32_t get_float32(self, float32_t key) except *:
        return i32_to_f32(self.get_int32(key))

    def __getitem__(self, key):
        if self.for_int:
            return self.get_int32(key)
        else:
            return self.get_float32(key)

    
    cpdef void discard(self, float32_t key) except *:
        cdef khint_t k
        k = kh_get_float32to32map(self.table, key)
        if k != self.table.n_buckets:
            kh_del_float32to32map(self.table, k)


    cdef Float32to32MapIterator get_iter(self):
        return Float32to32MapIterator(self)

    def __iter__(self):
        return self.get_iter()


### Iterator:
cdef class Float32to32MapIterator:

    cdef void __move(self) except *:
        while self.it<self.size and not kh_exist_float32to32map(self.parent.table, self.it):
              self.it+=1       

    cdef bint has_next(self) except *:
        return self.it != self.parent.table.n_buckets
        
    cdef float32to32_key_val_pair next(self) except *:
        cdef float32to32_key_val_pair result 
        result.key = self.parent.table.keys[self.it]
        result.val = self.parent.table.vals[self.it]
        self.it+=1#ensure at least one move!
        self.__move()
        return result


    def __cinit__(self, Float32to32Map parent):
        self.parent = parent
        self.size = parent.table.n_buckets
        #search the start:
        self.it = 0
        self.__move()

    def __next__(self):
        if self.has_next():
            return self.next()
        else:
            raise StopIteration

### Utils:

def Float32to32Map_from_int32_buffer(float32_t[:] keys, int32_t[:] vals, double size_hint=0.0):
    cdef Py_ssize_t n = len(keys)
    cdef Py_ssize_t b = len(vals)
    if b < n:
        n = b
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Float32to32Map(number_of_elements_hint=at_least_needed, for_int=True)
    cdef Py_ssize_t i
    for i in range(n):
        res.put_int32(keys[i], vals[i])
    return res

def Float32to32Map_from_float32_buffer(float32_t[:] keys, float32_t[:] vals,double size_hint=0.0):
    cdef Py_ssize_t n = len(keys)
    cdef Py_ssize_t b = len(vals)
    if b < n:
        n = b
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Float32to32Map(number_of_elements_hint=at_least_needed, for_int=False)
    cdef Py_ssize_t i
    for i in range(n):
        res.put_float32(keys[i], vals[i])
    return res
    







cdef extern from *:
    """
    // preprocessor creates needed struct-type and all function definitions

    // map with keys of type int32 -> und result int32
    // resulting typename: kh_int32to32map_t;
    KHASH_INIT(int32to32map, int32_t, int32_t, 1, kh_int_hash_func, kh_int_hash_equal)
  

    """
    pass







include "init_pyobjectset.pxi"
cdef extern from *:
    ctypedef struct kh_pyobjectset_t:
        khint_t n_buckets, size, n_occupied, upper_bound
        uint32_t *flags
        pyobject_t *keys
        #size_t *vals  //dummy

    kh_pyobjectset_t* kh_init_pyobjectset() nogil
    void kh_destroy_pyobjectset(kh_pyobjectset_t*) nogil
    void kh_clear_pyobjectset(kh_pyobjectset_t*) nogil
    khint_t kh_get_pyobjectset(kh_pyobjectset_t*, pyobject_t) nogil
    void kh_resize_pyobjectset(kh_pyobjectset_t*, khint_t) nogil
    khint_t kh_put_pyobjectset(kh_pyobjectset_t*, pyobject_t, int*) nogil
    void kh_del_pyobjectset(kh_pyobjectset_t*, khint_t) nogil

    #specializing "kh_exist"-macro 
    bint kh_exist_pyobjectset "kh_exist" (kh_pyobjectset_t*, khint_t) nogil


cdef class PyObjectSet:
    cdef kh_pyobjectset_t *table

    cdef bint contains(self, object key) except *
    cdef PyObjectSetIterator get_iter(self)
    cdef khint_t size(self) 
    cpdef void add(self, object key) except *
    cpdef void discard(self, object key) except *
    


cdef class PyObjectSetIterator:
    cdef khint_t   it
    cdef khint_t   size
    cdef PyObjectSet  parent

    cdef bint has_next(self) except *
    cdef object next(self)
    cdef void __move(self) except *




#
#
# Don't edit it, unless this is I_n_t_6_4_S_e_t header
#
# run sh all_from_XXX.sh to create it from bluepring - I_n_t_6_4_S_e_t

include "init_int64set.pxi"
cdef extern from *:

    ctypedef struct kh_int64set_t:
        khint_t n_buckets, size, n_occupied, upper_bound
        uint32_t *flags
        int64_t *keys
        #size_t *vals  //dummy

    kh_int64set_t* kh_init_int64set() nogil
    void kh_destroy_int64set(kh_int64set_t*) nogil
    void kh_clear_int64set(kh_int64set_t*) nogil
    khint_t kh_get_int64set(kh_int64set_t*, int64_t) nogil
    void kh_resize_int64set(kh_int64set_t*, khint_t) nogil
    khint_t kh_put_int64set(kh_int64set_t*, int64_t, int*) nogil
    void kh_del_int64set(kh_int64set_t*, khint_t) nogil

    #specializing "kh_exist"-macro 
    bint kh_exist_int64set "kh_exist" (kh_int64set_t*, khint_t) nogil


cdef class Int64Set:
    cdef kh_int64set_t *table

    cdef bint contains(self, int64_t key) except *
    cdef Int64SetIterator get_iter(self)
    cdef khint_t size(self) 
    cpdef void add(self, int64_t key) except *
    cpdef void discard(self, int64_t key) except *
    


cdef class Int64SetIterator:
    cdef khint_t   it
    cdef khint_t   size
    cdef Int64Set  parent

    cdef bint has_next(self) except *
    cdef int64_t next(self) except *
    cdef void __move(self) except *

cpdef Int64Set_from_buffer(int64_t[:] buf, double size_hint=*)





include "init_pyobjectmap.pxi"
cdef extern from *:

    ctypedef struct kh_pyobjectmap_t:
        khint_t n_buckets, size, n_occupied, upper_bound
        uint32_t *flags
        pyobject_t *keys
        pyobject_t *vals  

    kh_pyobjectmap_t* kh_init_pyobjectmap() nogil
    void kh_destroy_pyobjectmap(kh_pyobjectmap_t*) nogil
    void kh_clear_pyobjectmap(kh_pyobjectmap_t*) nogil
    khint_t kh_get_pyobjectmap(kh_pyobjectmap_t*, pyobject_t) nogil
    void kh_resize_pyobjectmap(kh_pyobjectmap_t*, khint_t) nogil
    khint_t kh_put_pyobjectmap(kh_pyobjectmap_t*, pyobject_t, int* result) nogil
    void kh_del_pyobjectmap(kh_pyobjectmap_t*, khint_t) nogil

    #specializing "kh_exist"-macro 
    bint kh_exist_pyobjectmap "kh_exist" (kh_pyobjectmap_t*, khint_t) nogil


cdef class PyObjectMap:
    cdef kh_pyobjectmap_t *table

    cdef bint contains(self, object key) except *
    cdef PyObjectMapIterator get_iter(self)
    cdef khint_t size(self) 
    cpdef void put_object(self, object key, object value) except *
    cpdef object get_object(self, object key)
    cpdef void discard(self, object key) except *
    

cdef struct pyobject_key_val_pair:
    pyobject_t key
    pyobject_t val


cdef class PyObjectMapIterator:
    cdef khint_t   it
    cdef khint_t   size
    cdef PyObjectMap  parent

    cdef bint has_next(self) except *
    cdef pyobject_key_val_pair next(self) except *
    cdef void __move(self) except *






cdef extern from *:
    """
    /*! @function
      @abstract     Instantiate a hash set containing 64-bit integer keys
      @param  name  Name of the hash table [symbol]
     */
    #define KHASH_SET_INIT_FLOAT64(name)										\
	    KHASH_INIT(name, khfloat64_t, char, 0, kh_float64_hash_func_0_NAN, kh_float64_hash_equal)

    //preprocessor creates needed struct-type and all function definitions 
    //set with keys of type float64 -> resulting typename: kh_float64set_t;
    KHASH_SET_INIT_FLOAT64(float64set)

    //TODO: add generated code:
    """
    pass




from cpython cimport buffer
from libc.stdlib cimport malloc, realloc, free


from .khashsets cimport khint_t, element_n_from_size_hint
from .khashsets cimport Int64Set, Int64Set_from_buffer, kh_exist_int64set
from .khashsets cimport Int32Set, Int32Set_from_buffer, kh_exist_int32set
from .khashsets cimport Float64Set, Float64Set_from_buffer, kh_exist_float64set
from .khashsets cimport Float32Set, Float32Set_from_buffer, kh_exist_float32set



cdef char *empty_buf=""

# responsible for freeing the data (so we are sure "right" free is called) and 
# can expose the data via buffer interface
# 
cdef class MemoryNanny:

    cdef void*       ptr
    cdef Py_ssize_t  n
    cdef Py_ssize_t  element_size
    cdef Py_ssize_t  buffer_lock_cnt
    cdef bytes       format

    def __cinit__(self):
        self.ptr = NULL
        self.n = 0
        self.element_size = 0
        self.buffer_lock_cnt = 0
        self.format = None 

    def __dealloc__(self):      
        free(self.ptr)
        self.ptr=NULL 

    def __getbuffer__(self, buffer.Py_buffer *view, int flags):
        #is input sane?
        if(view==NULL):
            raise BufferError("view==NULL argument is obsolete")


        #should never happen, just to be sure
        if NULL == self.ptr:
            view.buf = empty_buf
        else:
            view.buf = self.ptr

        view.obj = self; # increments ref count
       
        
        # size of the memory-block in bytes
        view.len = self.n * self.element_size

         
        # 0 or 1 possible as long as the same for all
        view.readonly = 0

        # should be original value, even if buffer.format is set to NULL
        view.itemsize = self.element_size;

        # format:
        if (flags & buffer.PyBUF_FORMAT) == buffer.PyBUF_FORMAT:
             view.format = self.format
        else:
             # NULL indicates that the buffer's data type has been cast to 'B'.
             # view->itemsize is the _previous_ itemsize. n * itemsize = len still holds at this
             # point. The equality calcsize(format) = itemsize does _not_ hold
             # from here on! */
             view.format = NULL

        # data is one-dimensional
        view.ndim = 1;
        

        view.shape = NULL
        view.strides = NULL
        view.suboffsets = NULL

        # no need for internal data
        view.internal = NULL

        self.buffer_lock_cnt+=1


    def __releasebuffer__(self, buffer.Py_buffer *view):
        self.buffer_lock_cnt-=1


    @staticmethod
    cdef MemoryNanny create_memory_nanny(void* ptr, Py_ssize_t  n, Py_ssize_t element_size, object format):
        cdef MemoryNanny nanny = MemoryNanny()
        nanny.ptr = ptr
        nanny.n   = n
        nanny.element_size = element_size
        nanny.format =  format
        return nanny


# TODO: reduce code duplication

cpdef unique_int64(int64_t[:] vals, double size_hint=0.0):
    cdef Int64Set s = Int64Set_from_buffer(vals, size_hint)
    
    # compress:
    cdef int64_t* mem = s.table.keys
    cdef khint_t i
    cdef khint_t current = 0
    for i in range(s.table.n_buckets):
        if kh_exist_int64set(s.table, i):
            mem[current] = mem[i]
            current += 1

    # take over the memory:
    s.table.keys = NULL
    
    # shrink to fit:
    mem = <int64_t*> realloc(mem, sizeof(int64_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(int64_t), b"q")


cpdef unique_stable_int64(int64_t[:] vals, double size_hint=0.0):
    # prepare 
    cdef Py_ssize_t n = len(vals)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Int64Set(number_of_elements_hint=at_least_needed)
    cdef int64_t* mem = <int64_t*> malloc(sizeof(int64_t)*n);
    
    # insert
    cdef khint_t current = 0
    cdef Py_ssize_t i
    cdef int64_t element
    for i in range(n):
        element = vals[i]
        res.add(element)
        if current != len(res):
            mem[current] = element
            current += 1
    
    # shrink to fit:
    mem = <int64_t*> realloc(mem, sizeof(int64_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(int64_t), b"q")


cpdef unique_int32(int32_t[:] vals, double size_hint=0.0):
    cdef Int32Set s = Int32Set_from_buffer(vals, size_hint)
    

    # compress:
    cdef int32_t* mem = s.table.keys
    cdef khint_t i
    cdef khint_t current = 0
    for i in range(s.table.n_buckets):
        if kh_exist_int32set(s.table, i):
            mem[current] = mem[i]
            current += 1

    # take over the memory:
    s.table.keys = NULL
    
    # shrink to fit:
    mem = <int32_t*> realloc(mem, sizeof(int32_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(int32_t), b"i")


cpdef unique_stable_int32(int32_t[:] vals, double size_hint=0.0):
    # prepare 
    cdef Py_ssize_t n = len(vals)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Int32Set(number_of_elements_hint=at_least_needed)
    cdef int32_t* mem = <int32_t*> malloc(sizeof(int32_t)*n);
    
    # insert
    cdef khint_t current = 0
    cdef Py_ssize_t i
    cdef int32_t element
    for i in range(n):
        element = vals[i]
        res.add(element)
        if current != len(res):
            mem[current] = element
            current += 1
    
    # shrink to fit:
    mem = <int32_t*> realloc(mem, sizeof(int32_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(int32_t), b"i")



cpdef unique_float64(float64_t[:] vals, double size_hint=0.0):
    cdef Float64Set s = Float64Set_from_buffer(vals, size_hint)
    
    # compress:
    cdef float64_t* mem = s.table.keys
    cdef khint_t i
    cdef khint_t current = 0
    for i in range(s.table.n_buckets):
        if kh_exist_float64set(s.table, i):
            mem[current] = mem[i]
            current += 1

    # take over the memory:
    s.table.keys = NULL
    
    # shrink to fit:
    mem = <float64_t*> realloc(mem, sizeof(float64_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(float64_t), b"d")


cpdef unique_stable_float64(float64_t[:] vals, double size_hint=0.0):
    # prepare 
    cdef Py_ssize_t n = len(vals)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Float64Set(number_of_elements_hint=at_least_needed)
    cdef float64_t* mem = <float64_t*> malloc(sizeof(float64_t)*n);
    
    # insert
    cdef khint_t current = 0
    cdef Py_ssize_t i
    cdef float64_t element
    for i in range(n):
        element = vals[i]
        res.add(element)
        if current != len(res):
            mem[current] = element
            current += 1
    
    # shrink to fit:
    mem = <float64_t*> realloc(mem, sizeof(float64_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(float64_t), b"d")


cpdef unique_float32(float32_t[:] vals, double size_hint=0.0):
    cdef Float32Set s = Float32Set_from_buffer(vals, size_hint)
    
    # compress:
    cdef float32_t* mem = s.table.keys
    cdef khint_t i
    cdef khint_t current = 0
    for i in range(s.table.n_buckets):
        if kh_exist_float32set(s.table, i):
            mem[current] = mem[i]
            current += 1

    # take over the memory:
    s.table.keys = NULL
    
    # shrink to fit:
    mem = <float32_t*> realloc(mem, sizeof(float32_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(float32_t), b"f")


cpdef unique_stable_float32(float32_t[:] vals, double size_hint=0.0):
    # prepare 
    cdef Py_ssize_t n = len(vals)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Float32Set(number_of_elements_hint=at_least_needed)
    cdef float32_t* mem = <float32_t*> malloc(sizeof(float32_t)*n);
    
    # insert
    cdef khint_t current = 0
    cdef Py_ssize_t i
    cdef float32_t element
    for i in range(n):
        element = vals[i]
        res.add(element)
        if current != len(res):
            mem[current] = element
            current += 1
    
    # shrink to fit:
    mem = <float32_t*> realloc(mem, sizeof(float32_t)*current);
    return MemoryNanny.create_memory_nanny(mem, current, sizeof(float32_t), b"f")




#
#
# Don't edit it, unless this is I_n_t_6_4_S_e_t implementation
#
# run sh all_from_XXX.sh to create it from blueprint - I_n_t_6_4_S_e_t
#
#

cdef class Float64Set:

    def __cinit__(self, *, number_of_elements_hint=None):
        """
        number_of_elements_hint - number of elements without the need of reallocation.
        """
        self.table = kh_init_float64set()
        if number_of_elements_hint is not None:    
            kh_resize_float64set(self.table, element_n_to_bucket_n(number_of_elements_hint))

    def __len__(self):
        return self.size()
  
    cdef khint_t size(self):
        return self.table.size
        

    def __dealloc__(self):
        if self.table is not NULL:
            kh_destroy_float64set(self.table)
            self.table = NULL

    def __contains__(self, float64_t key):
        return self.contains(key)


    cdef bint contains(self, float64_t key) except *:
        cdef khint_t k
        k = kh_get_float64set(self.table, key)
        return k != self.table.n_buckets


    cpdef void add(self, float64_t key) except *:
        cdef:
            khint_t k
            int ret = 0

        k = kh_put_float64set(self.table, key, &ret)
        self.table.keys[k] = key

    
    cpdef void discard(self, float64_t key) except *:
        cdef khint_t k
        k = kh_get_float64set(self.table, key)
        if k != self.table.n_buckets:
            kh_del_float64set(self.table, k)


    cdef Float64SetIterator get_iter(self):
        return Float64SetIterator(self)

    def __iter__(self):
        return self.get_iter()

    def get_state_info(self):
        return {"n_buckets" : self.table.n_buckets, 
                "n_occupied" : self.table.n_occupied, 
                "upper_bound" : self.table.upper_bound}


### Iterator:
cdef class Float64SetIterator:

    cdef void __move(self) except *:
        while self.it<self.size and not kh_exist_float64set(self.parent.table, self.it):
              self.it+=1       

    cdef bint has_next(self) except *:
        return self.it != self.parent.table.n_buckets
        
    cdef float64_t next(self) except *:
        cdef float64_t result = self.parent.table.keys[self.it]
        self.it+=1#ensure at least one move!
        self.__move()
        return result


    def __cinit__(self, Float64Set parent):
        self.parent = parent
        self.size = parent.table.n_buckets
        #search the start:
        self.it = 0
        self.__move()

    def __next__(self):
        if self.has_next():
            return self.next()
        else:
            raise StopIteration

### Utils:

def Float64Set_from(it):
    res=Float64Set()
    for i in it:
        res.add(i)
    return res

cpdef Float64Set_from_buffer(float64_t[:] buf, double size_hint=0.0):
    cdef Py_ssize_t n = len(buf)
    cdef Py_ssize_t at_least_needed = element_n_from_size_hint(<khint_t>n, size_hint)
    res=Float64Set(number_of_elements_hint=at_least_needed)
    cdef Py_ssize_t i
    for i in range(n):
        res.add(buf[i])
    return res
    


from libc.stdint cimport  uint8_t

cpdef isin_float64(float64_t[:] query, Float64Set db, uint8_t[:] result):
    cdef size_t i
    cdef size_t n=len(query)
    if n!=len(result):
        raise ValueError("Different sizes for query({n}) and result({m})".format(n=n, m=len(result)))
    for i in range(n):
        result[i]=db.contains(query[i])





from libc.stdint cimport  uint32_t,  int64_t,  int32_t
from cpython.object cimport PyObject

from .floatdef cimport float64_t, float32_t

### Common definitions:
ctypedef PyObject* pyobject_t

include "khash.pxi"


#utilities for int<->float
include "float_utils.pxi"
include "pyobject_utils.pxi"

# different implementation
include "maps/int64/int64to64map_header.pxi"
include "maps/int32/int32to32map_header.pxi"
include "maps/float64/float64to64map_header.pxi"
include "maps/float32/float32to32map_header.pxi"
include "maps/pyobject/pyobjectmap_header.pxi"

